import gym
from gym import error, spaces, utils
import numpy as np
from gym.utils import seeding



class StatefulNRM(gym.Env):
  # metadata = {'render.modes': ['human']}

  def __init__(self):
    self._seed = 1
    return

  def initialize(self, horizon, p, K, J, M, action_type, S0, action_price_map, XGenerator, Y_AX, nA = 2, seed =1):
    # Take in functional parametrization of X, and conditional response Y| A,X
    # M resource consumption matrix
    # product arrival rate Z
    # initial capacity S0
    # S0 is list of integer starting capacities for each product

    # dimension init parameters
    self.horizon = horizon
    self.p = p # context dimension
    self.J = J # number of resources
    self.K = K # number of products
    self.M = M
    self.XGenerator = XGenerator
    self.Y_AX = Y_AX
    self.num_envs = 1
    if action_type == 'discrete':
      self.nA = nA #
      self.action_space = spaces.Discrete(nA)
    else:
      self.action_space = spaces.Box(low=0,high=np.inf )
    # Product space of discrete state space and continuous context
    product_space = spaces.MultiDiscrete(S0)
    context_space = spaces.Box(low=-np.inf, high=np.inf, shape=(1,self.p) )
    product_arrival_space = spaces.Discrete(self.K)
    self.observation_space = spaces.Tuple( [product_space, context_space, product_arrival_space] )

    self.state = np.zeros(J+p+1)
    self.state[0:self.J] = S0
    k0 = self._productGenerator()
    X0 = self.XGenerator(0, k0)
    self.state[self.J:(self.J+p)] = X0
    self.state[-1] = k0
    self.S0 = S0
    self.action_price_map = action_price_map #k x nA map of, for each product, prices for each action
    # track timestep
    self.t = 0



    # Contextual demand model
    self.XGenerator = XGenerator
    self.Y_AX = Y_AX
    return

  def _price_from_a(self, k, action):
    k = int(k)
    if self.K > 1:
      return self.action_price_map[k, action]
    else:
      return self.action_price_map[action]

  def _productGenerator(self):
    return int(np.random.choice(self.K))
  def _step(self, action):

    obs = self.state
    S = self.state[0:self.J]
    k = int(self.state[-1]) # product index
    X = self.state[-(1+self.p):-1]

    # transition
    price = self._price_from_a(k, action)

    Y = self.Y_AX(price, X, k) # purchase/nopurchase outcome
    feasible = np.all(S >= self.M[k,:])
    S_prime = S - feasible * Y * self.M[k,:] # subtract resource consumption if purchase
    reward = self._get_reward(Y, k, action)

    # generate product request
    # generate context given product request
    self.t += 1
    k_prime = self._productGenerator()
    X_prime = self.XGenerator(self.t, k_prime)

    # (This generated data is used for the next state).
    self.state = np.hstack([S_prime, X_prime, k_prime])

    # check if unable to make any products
    episode_over = np.all( [S <= self.M[k,:] for k in range(self.K) ])

    return self.state, reward, episode_over, {}

  def _get_reward(self, Y, k, action):
    S = self.state[0:self.J]
    # feasible if we can make the product
    feasible = np.all(S >= self.M[k,:])
    return feasible * Y * self._price_from_a(k,action)

  def _reset(self):
    self.state[0:self.J] = self.S0
    k0 = self._productGenerator()
    X0 = self.XGenerator(0, k0)
    self.state[self.J:(self.J+self.p)] = X0
    self.state[-1] = k0
  def _render(self):
    return
  def close(self):
    return

