import numpy as np
import gurobipy as gp
import random

# class Gssprime:
# # initialize with a mapper function
# # mapper: takes (s,sprime,y) or (s,sprime) query and returns 0,1 feasible
#   def __init__(self, mapper, Ss, Yy,verbose=True):
#     self.mapper = mapper
#     self.Ss = Ss # 1-d state representation
#     self.Yy = Yy
#     self.map_ = dict()
#
#     if verbose:
#     # precompute sparsity pattern on initialize
#     # for every state: generate tuples of (next state, y)
#         for s in Ss:
#             next_transitions = [(sprime,y) for sprime in Ss for y in Yy if self.mapper(s,sprime,y) != 0 ]
#             self.map_[s] = next_transitions # next transitions from s under y realization
#
#   def get(self, s, sprime, y = None):
#     if y is not None:
#         # get index
#         return self.mapper(s, sprime,y)
#     else:
#         # return row
#         return self.mapper(s, sprime)

class Gssprime_multiarrival:
# initialize with a mapper function
# mapper: takes (s,sprime,y) or (s,sprime) query and returns 0,1 feasible
  def __init__(self, mapper, Ss, Yy, nrm_config, verbose=True):
    self.mapper = mapper
    self.Ss = Ss # 1-d state representation
    self.Yy = Yy
    self.map_ = dict() # nested dictionary if verbose
    self.K = nrm_config.K
    self.M = nrm_config.M



    if verbose:
        for k in range(self.K):
            self.map_[k] = dict()
        # precompute sparsity pattern on initialize
        # for every state: generate tuples of (next state, y)
            for s in Ss:
                next_transitions = [(sprime,y) for sprime in Ss for y in Yy if self.mapper(s,sprime,y,k) != 0 ]
                self.map_[k][s] = next_transitions # next transitions from s under y realization

  def get(self, s, sprime, k, y = None):
    if y is not None:
        # get index
        return self.mapper(s, sprime,y,k)
    else:
        # return row
        return self.mapper(s, sprime,k)

class Gssprime_singlearrival:
# initialize with a mapper function
# mapper: takes (s,sprime,y) or (s,sprime) query and returns 0,1 feasible
  def __init__(self, mapper, Ss, Yy, nrm_config, verbose=True):
    self.mapper = mapper
    self.Ss = Ss # 1-d state representation
    self.Yy = Yy
    self.map_ = dict() # nested dictionary if verbose
    if verbose:
        # precompute sparsity pattern on initialize
        # for every state: generate tuples of (next state, y)
        for s in Ss:
            next_transitions = [(sprime,y) for sprime in Ss for y in Yy if self.mapper(s,sprime,y) != 0 ]
            self.map_[s] = next_transitions # next transitions from s under y realization

  def get(self, s, sprime, y = None):
    if y is not None:
        # get index
        return self.mapper(s, sprime,y)
    else:
        # return row
        return self.mapper(s, sprime)



# speedup with linexpr

import scipy.sparse as sp
from numpy import load, save, asarray
from multiprocessing import Pool
from functools import partial


def PLRALP_singlearrival(config_,margSs, # problem parameters
            S_, X_, A_, Y_,                   # data
            score, e_AX, Y_AX ):                   # estimation strategy and nuisances

    # In this function, margSs refers to _marginal state margSs_
    # dS is tuple of state dimension
    # margSs is iterable of state variables, Yy is iterable of Y values
    # dynamic_Gs assumes G_s is accessible in constant lookup
    # X_ is [ N,T,p ] array
    # assume score is a function of (X,A,Y,y,e_AX,Y_AX)
    dS = len(margSs) # one dimensional representation
    nA = config_.nA
    [N,horizon,p] = X_.shape
    # add an intercept
    X_augmented = np.ones([N,horizon,p+1])
    X_augmented[:,:,0:p] = X_
    X_noint = X_
    X_ = X_augmented
    p=int(p+1)

    # Create variables  # dimension of theta
    m = gp.Model()
    theta = m.addVars(horizon+1,dS,p, vtype=gp.GRB.CONTINUOUS, lb=-1000,ub=1000, name="theta") # theta indexed by scalar version of state
    scalar_Ss = dict(zip(margSs, [ind for ind, s in enumerate(margSs)]))

    # Set objective
    X0mean = np.mean(X_[:,0,:],axis=0)
    print(X0mean)
    m.setObjective( gp.quicksum(
        X0mean[j] * theta[0,s_,j]
        for j in range(p)
        for s_ in range(dS)
    ) , gp.GRB.MINIMIZE)


    for t in range(horizon):
        print(t,'t')
        for a in range(nA):
            print(a,'a')
            for s in range(dS):
                print(s, 's')
                if s > 1:
                # sum over sparse lookup
                    if t < horizon-1:
                        m.addConstrs(
                            gp.quicksum(
                                1.0/N*np.abs(X_[i,t,p_])
                                *  # average
                                (
                                gp.quicksum(X_[i,t,j]*theta[t,s,j] for j in range(p) )
                                -
                                gp.quicksum( # transition Y = 1
                                              score(X_noint[i,t,:],A_[i,t],a, Y_[i,t],1,e_AX,Y_AX) *
                                              config_.action_price_map[a]+
                                              score(X_noint[i,t,:],A_[i,t],a, Y_[i,t],1,e_AX,Y_AX) *theta[t+1,s-1,j__]*np.mean(X_[:,t+1,j__],axis=0)
                                                             for j__ in range(p)  # tilde V_t+1
                                            )
                                - gp.quicksum(  score(X_noint[i,t,:],A_[i,t],a, Y_[i,t],0,e_AX,Y_AX) *theta[t+1,s,j__]*np.mean(X_[:,t+1,j__],axis=0) for j__ in range(p))
                                              )
                                  for i in range(N))    >= 0
                                       for p_ in range(p))
                    else:
                        print(N,'N')
                        print(p,'p')
                        m.addConstrs(
                            gp.quicksum(
                                1.0 / N * np.abs(X_[i, t, p_])
                                *  # average
                                (gp.quicksum(X_[i, t, j] * theta[t, s, j] for j in range(p))
                                        - score(X_noint[i, t, :], A_[i, t], a, Y_[i, t], 1, e_AX, Y_AX) *config_.action_price_map[a] # transition Y = 1
                                )
                                for i in range(N) ) >= 0
                            for p_ in range(p) )
    m.addConstrs( theta[horizon,s,j] == 0 for s in range(dS) for j in range(p))
# Optimize model
    m.optimize()
    return [m, theta]


def PLRALP_singlearrival_Xconstrs(config_,margSs, # problem parameters
            S_, X_, A_, Y_,                   # data
            score, e_AX, Y_AX ):                   # estimation strategy and nuisances
    '''
    add sampled constraints to marginal constraints
    '''

    # In this function, margSs refers to _marginal state margSs_
    # dS is tuple of state dimension
    # margSs is iterable of state variables, Yy is iterable of Y values
    # dynamic_Gs assumes G_s is accessible in constant lookup
    # X_ is [ N,T,p ] array
    # assume score is a function of (X,A,Y,y,e_AX,Y_AX)
    dS = len(margSs) # one dimensional representation
    nA = config_.nA
    [N,horizon,p] = X_.shape
    # add an intercept
    X_augmented = np.ones([N,horizon,p+1])
    X_augmented[:,:,0:p] = X_
    X_noint = X_
    X_ = X_augmented
    p=int(p+1)

    # Create variables  # dimension of theta
    m = gp.Model()
    theta = m.addVars(horizon+1,dS,p, vtype=gp.GRB.CONTINUOUS, lb=-300,ub=300, name="theta") # theta indexed by scalar version of state
    scalar_Ss = dict(zip(margSs, [ind for ind, s in enumerate(margSs)]))

    # Set objective
    X0mean = np.mean(X_[:,0,:],axis=0)
    print(X0mean)
    m.setObjective( gp.quicksum(
        X0mean[j] * theta[0,s_,j]
        for j in range(p)
        for s_ in range(dS)
    ) , gp.GRB.MINIMIZE)

    N_CONSTRS = 50
    CONSTR_INDS = random.sample(range(0,N), N_CONSTRS)

    for t in range(horizon):
        print(t,'t')
        for a in range(nA):
            print(a,'a')
            for s in range(dS):
                print(s, 's')
                if s > 1:
                # sum over sparse lookup
                    if t < horizon-1:
                        m.addConstrs(
                            gp.quicksum(
                                1.0/N*np.abs(X_[i,t,p_])
                                *  # average
                                (
                                gp.quicksum(X_[i,t,j]*theta[t,s,j] for j in range(p) )
                                -
                                gp.quicksum( # transition Y = 1
                                              score(X_noint[i,t,:],A_[i,t],a, Y_[i,t],1,e_AX,Y_AX) *
                                              config_.action_price_map[a]+
                                              score(X_noint[i,t,:],A_[i,t],a, Y_[i,t],1,e_AX,Y_AX) *theta[t+1,s-1,j__]*np.mean(X_[:,t+1,j__],axis=0)
                                                             for j__ in range(p)  # tilde V_t+1
                                            )
                                - gp.quicksum(  score(X_noint[i,t,:],A_[i,t],a, Y_[i,t],0,e_AX,Y_AX) *theta[t+1,s,j__]*np.mean(X_[:,t+1,j__],axis=0) for j__ in range(p))
                                              )
                                  for i in range(N))    >= 0
                                       for p_ in range(p))

                    else:
                        print(N,'N')
                        print(p,'p')
                        m.addConstrs(
                            gp.quicksum(
                                1.0 / N * np.abs(X_[i, t, p_])
                                *  # average
                                (gp.quicksum(X_[i, t, j] * theta[t, s, j] for j in range(p))
                                        - score(X_noint[i, t, :], A_[i, t], a, Y_[i, t], 1, e_AX, Y_AX) *config_.action_price_map[a] # transition Y = 1
                                )
                                for i in range(N) ) >= 0
                            for p_ in range(p) )

                        m.addConstrs(
                                gp.quicksum(X_[_i,t,j]*theta[t,s,j] for j in range(p) )
                                 - score(X_noint[_i, t, :], A_[_i, t], a, Y_[_i, t], 1, e_AX, Y_AX) *config_.action_price_map[a]
                                      >= 0
                                       for _i in CONSTR_INDS)
    m.addConstrs( theta[horizon,s,j] == 0 for s in range(dS) for j in range(p))
# Optimize model
    m.optimize()
    return [m, theta]

def PLRALP_singlearrival_sampleXconstrs(config_,margSs, # problem parameters
            S_, X_, A_, Y_,                   # data
            score, e_AX, Y_AX ):                   # estimation strategy and nuisances

    # In this function, margSs refers to _marginal state margSs_
    # dS is tuple of state dimension
    # margSs is iterable of state variables, Yy is iterable of Y values
    # dynamic_Gs assumes G_s is accessible in constant lookup
    # X_ is [ N,T,p ] array
    # assume score is a function of (X,A,Y,y,e_AX,Y_AX)
    dS = len(margSs) # one dimensional representation
    nA = config_.nA
    [N,horizon,p] = X_.shape
    # add an intercept
    X_augmented = np.ones([N,horizon,p+1])
    X_augmented[:,:,0:p] = X_
    X_noint = X_
    X_ = X_augmented
    p=int(p+1)

    # Create variables  # dimension of theta
    m = gp.Model()
    theta = m.addVars(horizon+1,dS,p, vtype=gp.GRB.CONTINUOUS, lb=0,ub=300, name="theta") # theta indexed by scalar version of state
    scalar_Ss = dict(zip(margSs, [ind for ind, s in enumerate(margSs)]))

    # Set objective
    X0mean = np.mean(X_[:,0,:],axis=0)
    print(X0mean)
    m.setObjective( gp.quicksum(
        X0mean[j] * theta[0,s_,j]
        for j in range(p)
        for s_ in range(dS)
    ) , gp.GRB.MINIMIZE)

    N_CONSTRS = 400
    CONSTR_INDS = random.sample(range(0,N), N_CONSTRS)

    for t in range(horizon):
        print(t,'t')
        for a in range(nA):
            print(a,'a')
            for s in range(dS):
                print(s, 's')
                if s > 1:
                # sum over sparse lookup
                    if t < horizon-1:
                        m.addConstrs(
                                gp.quicksum(X_[_i,t,p_]*theta[t,s,p_] for p_ in range(p))
                                -
                                gp.quicksum(score(X_noint[_i,t,:],A_[_i,t],a, Y_[_i,t],1,e_AX,Y_AX) *theta[t+1,s-1,j__]*np.mean(X_[:,t+1,j__],axis=0)
                                                             for j__ in range(p)  # tilde V_t+1
                                            )# transition Y = 1
                                - gp.quicksum(  score(X_noint[_i,t,:],A_[_i,t],a, Y_[_i,t],0,e_AX,Y_AX) *theta[t+1,s,j__]*np.mean(X_[:,t+1,j__],axis=0) for j__ in range(p))
                                     >= score(X_noint[_i,t,:],A_[_i,t],a, Y_[_i,t],1,e_AX,Y_AX) *config_.action_price_map[a]
                                       for _i in CONSTR_INDS )

                    else:
                        # m.addConstrs(
                        #     gp.quicksum(
                        #         1.0 / N * np.abs(X_[i, t, p_])
                        #         *  # average
                        #         (gp.quicksum(X_[i, t, j] * theta[t, s, j] for j in range(p))
                        #                 - score(X_noint[i, t, :], A_[i, t], a, Y_[i, t], 1, e_AX, Y_AX) *config_.action_price_map[a] # transition Y = 1
                        #         )
                        #         for i in range(N) ) >= 0
                        #     for p_ in range(p) )

                        m.addConstrs(
                                gp.quicksum(X_[_i,t,j]*theta[t,s,j] for j in range(p) )
                                 - score(X_noint[_i, t, :], A_[_i, t], a, Y_[_i, t], 1, e_AX, Y_AX) *config_.action_price_map[a]
                                      >= 0
                                       for _i in CONSTR_INDS)
    m.addConstrs( theta[horizon,s,j] == 0 for s in range(dS) for j in range(p))
# Optimize model
    m.optimize()
    return [m, theta]

def PLRALP_singlearrival_V_sx(config_,margSs, # problem parameters
            S_, X_, A_, Y_,                   # data
            score, e_AX, Y_AX ):                   # estimation strategy and nuisances

    # In this function, margSs refers to _marginal state margSs_
    # dS is tuple of state dimension
    # margSs is iterable of state variables, Yy is iterable of Y values
    # dynamic_Gs assumes G_s is accessible in constant lookup
    # X_ is [ N,T,p ] array
    # assume score is a function of (X,A,Y,y,e_AX,Y_AX)
    dS = len(margSs) # one dimensional representation
    nA = config_.nA
    [N,horizon,p] = X_.shape
    # add an intercept
    X_augmented = np.ones([N,horizon,p+1])
    X_augmented[:,:,0:p] = X_
    X_noint = X_
    X_ = X_augmented
    p=int(p+1)

    # Create variables  # dimension of theta
    m = gp.Model()
    # theta = m.addVars(horizon+1,dS,p, vtype=gp.GRB.CONTINUOUS, lb=0,ub=300, name="theta") # theta indexed by scalar version of state

    N_CONSTRS = 400
    CONSTR_INDS = random.sample(range(0,N), N_CONSTRS)
    V = m.addVars(horizon + 1, dS, N_CONSTRS, vtype=gp.GRB.CONTINUOUS, lb=0, ub=300, name="theta")
    # Set objective
    X0mean = np.mean(X_[:,0,:],axis=0)
    print(X0mean)
    m.setObjective( gp.quicksum(
        V[0,s,i] for s in range(dS) for i in range(N_CONSTRS)
    ) , gp.GRB.MINIMIZE)


    for t in range(horizon):
        print(t,'t')
        for a in range(nA):
            print(a,'a')
            for s in range(dS):
                print(s, 's')
                if s > 0:
                # sum over sparse lookup
                    if t < horizon-1:
                        m.addConstrs(
                            (V[t,s,_i] # transition Y = 1
                                - score(X_noint[_i,t,:],A_[_i,t],a, Y_[_i,t],1,e_AX,Y_AX) * gp.quicksum(1.0/N_CONSTRS * V[t+1,s-1,__i] for  __i in range(N_CONSTRS)) # tilde V_t+1
                                - score(X_noint[_i,t,:],A_[_i,t],a, Y_[_i,t],0,e_AX,Y_AX) * gp.quicksum(1.0/N_CONSTRS * V[t+1,s,__i] for  __i in range(N_CONSTRS))
                                     >= score(X_noint[_i,t,:],A_[_i,t],a, Y_[_i,t],1,e_AX,Y_AX) *config_.action_price_map[a]
                                       for _i in range(N_CONSTRS) ), name=str(t)+','+str(a)+','+str(s) )

                    else:
                        # m.addConstrs(
                        #     gp.quicksum(
                        #         1.0 / N * np.abs(X_[i, t, p_])
                        #         *  # average
                        #         (gp.quicksum(X_[i, t, j] * theta[t, s, j] for j in range(p))
                        #                 - score(X_noint[i, t, :], A_[i, t], a, Y_[i, t], 1, e_AX, Y_AX) *config_.action_price_map[a] # transition Y = 1
                        #         )
                        #         for i in range(N) ) >= 0
                        #     for p_ in range(p) )

                        m.addConstrs(
                            (V[t,s,_i]- score(X_noint[_i, t, :], A_[_i, t], a, Y_[_i, t], 1, e_AX, Y_AX) *config_.action_price_map[a]
                                      >= 0
                                       for _i in range(N_CONSTRS)), name = str(t)+','+str(a)+','+str(s))
    m.addConstrs( V[horizon,s,i] == 0 for s in range(dS) for i in range(N_CONSTRS))
# Optimize model
    m.optimize()
    return [m, V]

def PLRALP_singlearrival_V_sx_dualfnapprox(config_,margSs, # problem parameters
            S_, X_, A_, Y_,                   # data
            score, e_AX, Y_AX ):                   # estimation strategy and nuisances

    # In this function, margSs refers to _marginal state margSs_
    # dS is tuple of state dimension
    # margSs is iterable of state variables, Yy is iterable of Y values
    # dynamic_Gs assumes G_s is accessible in constant lookup
    # X_ is [ N,T,p ] array
    # assume score is a function of (X,A,Y,y,e_AX,Y_AX)
    dS = len(margSs) # one dimensional representation
    nA = config_.nA
    [N,horizon,p] = X_.shape
    # add an intercept
    X_augmented = np.ones([N,horizon,p+1]); X_augmented[:,:,0:p] = X_; X_noint = X_;X_ = X_augmented;p=int(p+1)

    # Create variables  # dimension of theta
    m = gp.Model()
    N_CONSTRS = 400
    V = m.addVars(horizon + 1, dS, N_CONSTRS, vtype=gp.GRB.CONTINUOUS, lb=0, ub=300, name="theta")
    # Set objective
    X0mean = np.mean(X_[:,0,:],axis=0)
    print(X0mean)
    m.setObjective( gp.quicksum(
        V[0,s,i] for s in range(dS) for i in range(N_CONSTRS)
    ) , gp.GRB.MINIMIZE)


    for t in range(horizon):
        print(t,'t')
        for a in range(nA):
            print(a,'a')
            for s in range(dS):
                print(s, 's')
                if s > 0:
                # sum over sparse lookup
                    if t < horizon-1:

                        m.addConstrs(
                            (gp.quicksum(
                                1.0/N_CONSTRS*X_[i,t,p_]
                                *  # average
                                (
                                V[t,s,i] # transition Y = 1
                                -score(X_noint[i,t,:],A_[i,t],a, Y_[i,t],1,e_AX,Y_AX) *config_.action_price_map[a]+
                                 score(X_noint[i,t,:],A_[i,t],a, Y_[i,t],1,e_AX,Y_AX) *gp.quicksum(1.0/N_CONSTRS * V[t+1,s-1,__i] for  __i in range(N_CONSTRS))
                                            )
                                -  score(X_noint[i,t,:],A_[i,t],a, Y_[i,t],0,e_AX,Y_AX) *gp.quicksum(1.0/N_CONSTRS * V[t+1,s-1,__i] for  __i in range(N_CONSTRS))
                                  for i in range(N_CONSTRS))    >= 0
                                       for p_ in range(p)), name = str(t)+','+str(a)+','+str(s))
                    else:
                        m.addConstrs((
                                V[t,s,_i]- score(X_noint[_i, t, :], A_[_i, t], a, Y_[_i, t], 1, e_AX, Y_AX) *config_.action_price_map[a]
                                      >= 0
                                       for _i in range(N_CONSTRS)), name = str(t)+','+str(a)+','+str(s))
    m.addConstrs( V[horizon,s,i] == 0 for s in range(dS) for i in range(N_CONSTRS))
# Optimize model
    m.optimize()
    return [m, V]
#

#
#
# def PLRALP_dynamic_Gs_multiarrival(config_,margSs, # problem parameters
#             S_, X_, A_, Y_, K_,                   # data
#             G_s, R,                                # problem mapping
#             score, e_AX, Y_AX ):                   # estimation strategy and nuisances
#
#     # In this function, margSs refers to _marginal state margSs_
#
#     # dS is tuple of state dimension
#     # margSs is iterable of state variables, Yy is iterable of Y values
#     # dynamic_Gs assumes G_s is accessible in constant lookup
#     # X_ is [ N,T,p ] array
#     # assume score is a function of (X,A,Y,y,e_AX,Y_AX)
#     # load from config
#     dS = len(margSs) # one dimensional representation
#     nA = config_.nA
#     [N,horizon,p] = X_.shape
#     K = config_.K
#     f = config_.arrival_type_dens
#     scalar_Ss = dict(zip(margSs, [ind for ind, s in enumerate(margSs)]))
#     # Create variables  # dimension of theta
#     m = gp.Model()
#     theta_dim = tuple([horizon] + [dS] + [K] + [p])
#     theta = m.addVars(horizon,dS,K,p, vtype=gp.GRB.CONTINUOUS, lb=0, name="theta") # theta indexed by scalar version of state
#     # Set objective
#     X0mean = np.mean(X_[:,0,:],axis=0)
#     print(X0mean)
#     m.setObjective( gp.quicksum(
#         X0mean[j] * theta[0,scalar_Ss[s_],k,j]
#         for j in range(p)
#         for k in range(K)
#         for s_ in margSs
#     ) , gp.GRB.MINIMIZE)
#
#
#     for t in range(horizon):
#         print(t)
#         for a in range(nA):
#             print(a,'a')
#             for s in margSs:
#                 for k in range(K):
#                     print(s)
#                     # sum over sparse lookup
#                     m.addConstrs( (1.0/N*
#                         gp.quicksum(
#                             np.abs(X_[i,t,p_]))
#                             *  # average
#                             (
#                             gp.quicksum(f(t, k__)*X_[i,t,j]*theta[t,scalar_Ss[s],k__,j] for j in range(p) for k__ in range(K))
#                             -
#                             gp.quicksum(  f(t, k)*
#                                           score(X_[i,t,:],A_[i,t],a, Y_[i,t],y,e_AX,Y_AX,K_[i,t], k) *
#                                           (R(s,a,sprime,k)+
#                                             gp.quicksum( f(t+1,k__) *
#                                                          theta[t+1,scalar_Ss[sprime],k__,j__]*np.mean(X_[:,t+1,j__],axis=0)
#                                                          for j__ in range(p) for k__ in range(K)) # tilde V_t+1
#                                         )
#                                           for k in range(K) for (sprime,y) in G_s.map_[k][s] # next states
#                                           )
#                              )) for i in range(N)
#                                    for p_ in range(p))
#                                   >= 0 , "alp")
#     # Optimize model
#     m.optimize()
#     return [m, theta]


def preprocess_X(X_,add_intercept=True):
    [N, horizon, p] = X_.shape
    Xouters = dict()
    if add_intercept:
        X_intercept_augmented = np.ones([N,horizon,p+1])
        X_intercept_augmented[:,:,0:p] = X_
        X_ = X_intercept_augmented

    for t in range(horizon):
        Xouters[t,t] = (1.0/N**2)*(X_[:,t,:].T @ np.abs(X_[:,t,:]))
        if t < horizon-1:
            Xouters[t,t+1] = (1.0/N**2)* (X_[:,t+1,:].T @ np.abs(X_[:,t,:]))

    return Xouters

def return_values(s,t,a,k,Xouters,
    config_, margSs, X_, A_, Y_, K_,  # data
    G_s, R,  # problem mapping
    score, e_AX, Y_AX ):
    # assume xouters, scores given
        # row by row construction
        # return list of i,j,v for a^T x < b
        # this state

    dS = len(margSs)  # one dimensional representation
    nA = config_.nA
    [N, horizon, p] = X_.shape
    Xp = p
    p = p+1 # add intercept
    X_intercept_augmented = np.ones([N, horizon, p ])
    X_intercept_augmented[:, :, 0:(p-1)] = X_

    K = config_.K
    f = config_.arrival_type_dens
    scalar_Ss = dict(zip(margSs, [ind for ind, s in enumerate(margSs)]))

    A_rowshape=(horizon,dS,nA,p,K)
    A_colshape=(horizon,dS,p,K)

    #t (V(s,x)) adding a theta-sized block of constraints (p many coefficients for every row)
    # one array for each dimension
    # [t,s,a,rowj,k]
    if t < horizon-1:
        is_ = np.ravel_multi_index( (np.ones(p**2,dtype=int)*t,np.ones(p**2,dtype=int)*scalar_Ss[s],np.ones(p**2,dtype=int)*a, [rowj for rowj in range(p) for colj in range(p) ],np.ones(p**2,dtype=int)*k), A_rowshape ) # row index
        # indices into theta
        js_ = np.ravel_multi_index((np.ones(p**2,dtype=int)*t,np.ones(p**2,dtype=int)*scalar_Ss[s], [colj for rowj in range(p) for colj in range(p) ],np.ones(p**2,dtype=int)*k), A_colshape)

        vals = [ f(t,k) * Xouters[(t,t)][rowj,colj] for rowj in range(p) for colj in range(p) ]
        # expectation over next time step
        print(K)

        for (sprime,y) in G_s.map_[k][s]:
            # same constraint row
            newis = np.ravel_multi_index( (np.ones((p ** 2) * K,dtype=int) * t, np.ones((p ** 2) * K,dtype=int) * scalar_Ss[s],
                                          np.ones((p ** 2) * K,dtype=int) * a,
                                        [rowj for rowj in range(p) for colj in range(p) for k__ in range(K)], np.ones((p ** 2) * K,dtype=int) * k),
                                       A_rowshape)
            is_ = np.hstack([is_, newis])  # row index
            newjs = np.ravel_multi_index( (np.ones((p ** 2) * K,dtype=int) * (t+1), np.ones((p ** 2) * K,dtype=int) * scalar_Ss[sprime],
                                          [colj for rowj in range(p) for colj in range(p) for k__ in range(K)],
                                          [k__ for rowj in range(p) for colj in range(p) for k__ in range(K)]),A_colshape)
            js_ = np.hstack([js_,newjs])
            # np.array([[t+1, scalar_Ss[sprime], colj, k__] for rowj in range(p) for colj in range(p) for k__ in range(K) ])
            Xtplus_mean = np.mean(X_intercept_augmented[:,t+1,:],axis=0)
            scores = score(X_[:,t,:],A_[:,t],a,Y_[:,t],y,e_AX,Y_AX,K,k)
            print(scores)
            newvals = [-1*f(t, k__) * np.mean(np.multiply(scores,X_intercept_augmented[:,t,rowj])) * Xtplus_mean[colj] for rowj in range(p) for colj in range(p) for k__ in range(K)]
            vals = np.hstack([vals,newvals])
        b_inds_ = np.ravel_multi_index((np.ones(p,dtype=int)*t,np.ones(p,dtype=int)*scalar_Ss[s],np.ones(p,dtype=int)*a, np.array([rowj for rowj in range(p)]),np.ones(p,dtype=int)*k), A_rowshape ) # row index,A_rowshape)
        print( [R(s,a,sprime,k) for (sprime,y) in G_s.map_[k][s]])
        bs_ = [ sum([np.mean(np.multiply(scores,X_intercept_augmented[:,t,rowj]))*R(s,a,sprime,k) for (sprime,y) in G_s.map_[k][s]]) for rowj in range(p) ]

    # Enforce V_T = 0 manually in the linear program

    return [is_, js_, vals, b_inds_, bs_]


def build_matrix(A_matrix,bs_,
    config_, margSs, quiet = True):
    '''

    '''
    # assume xouters, scores given
        # row by row construction
        # return list of i,j,v for a^T x < b
        # this state

    dS = len(margSs)  # one dimensional representation
    nA = config_.nA
    horizon = config_.horizon; p = config_.p
    p = p+1 # intercept
    K = config_.K
    f = config_.arrival_type_dens
    scalar_Ss = dict(zip(margSs, [ind for ind, s in enumerate(margSs)]))
    A_rowshape=(horizon,dS,nA,p,K); A_colshape=(horizon,dS,p,K)
    m = gp.Model()
    if quiet: m.setParam("OutputFlag", 0)
    thetas = m.addMVar(shape=horizon * dS * p * K, vtype=gp.GRB.CONTINUOUS, lb=0, ub = 1000)
    m.addConstr(A_matrix @ thetas >= bs_)
    print('added constraints')
    horizon_indices = np.ravel_multi_index((np.ones(dS*p*K,dtype=int)*(horizon-1),
                    [s for s in range(dS) for j_ in range(p) for k_ in range(K)],
                    [j_ for s in range(dS) for j_ in range(p) for k_ in range(K)],
                    [k_ for s in range(dS) for j_ in range(p) for k_ in range(K)]),
                    A_colshape)
    m.addConstrs( thetas[i] == 0 for i in horizon_indices )
    initial_indices = np.ravel_multi_index((np.ones(dS*p*K,dtype=int)*(0),[s for s in range(dS) for j_ in range(p) for k_ in range(K)],[j_ for s in range(dS) for j_ in range(p) for k_ in range(K)],[k_ for s in range(dS) for j_ in range(p) for k_ in range(K)]),
                    A_colshape)
    obj = np.zeros(horizon*dS*p*K)
    obj[initial_indices] = 1
    m.setObjective( obj @ thetas,gp.GRB.MINIMIZE)
    m.optimize()
    return m


def backwards_induction_Gs_multiarrival(config_,margSs, # problem parameters
            S_, X_, A_, Y_, K_,                   # data
            X_outers,
            G_s, R,                                # problem mapping
            score, e_AX, Y_AX ):                   # estimation strategy and nuisances
    '''
    Solve for policy by backwards induction
    # dS is tuple of state dimension
    # margSs is iterable of state variables, Yy is iterable of Y values
    # dynamic_Gs assumes G_s is accessible in constant lookup
    # X_ is [ N,T,p ] array
    # assume score is a function of (X,A,Y,y,e_AX,Y_AX)
    # load from config
    :return:
    '''
    # In this function, margSs refers to _marginal state margSs_


    dS = len(margSs) # one dimensional representation
    nA = config_.nA
    [N,horizon,p] = X_.shape
    K = config_.K
    dimS = config_.dimS # tuple dimension of state space
    f = config_.arrival_type_dens
    gamma = config_.gamma
    scalar_Ss = dict(zip(margSs, [ind for ind, s in enumerate(margSs)]))
    # Create variables  # dimension of theta
    m = gp.Model()
    theta_dim = tuple([horizon] + [dS] + [K] + [p])
    theta = m.addVars(horizon,dS,K,p, vtype=gp.GRB.CONTINUOUS, lb=0, name="theta") # theta indexed by scalar version of state
    # Set objective
    X0mean = np.mean(X_[:,0,:],axis=0)
    print(X0mean)
    m.setObjective( gp.quicksum(
        X0mean[j] * theta[0,scalar_Ss[s_],k,j]
        for j in range(p)
        for k in range(K)
        for s_ in margSs
    ) , gp.GRB.MINIMIZE)

    V = np.zeros([k,horizon]+dimS)
    for t in reversed(range(horizon)):
        for k in range(K):
            for s in margSs:
                # sum over sparse lookup# policy evaluation
                scores = np.zeros([N,nA])
                for (sprime, y) in G_s.map_[k][s]:
                    for a in range(nA):
                        scores[:,a] = score(X_[:, t, :], A_[:, t], a, Y_[:, t], y, e_AX, Y_AX, K, k)*(R(s,a,sprime,k) + V[t+1,sprime])
                        # range over x
                    best_a = np.argmax(scores,axis=1)

                    V[(k,t)+s] += np.mean( scores[:,best_a] )

    return V
#
# ### Code adapted from
# ## https://github.com/Self-guided-Approximate-Linear-Programs/Self-guided-ALPs-and-Related-Benchmarks/blob/master/ALP/ALPMaker.py
# def PLRALP_generate_scipy_sparse_matrix(config_,margSs, # problem parameters
#             S_, X_, A_, Y_, K_,                   # data
#             G_s, R,                                # problem mapping
#             score, e_AX, Y_AX ):                   # estimation strategy and nuisances
#
# # matrix is H*|A|*|S|*p x H*|S|*p
#
#     dS = len(margSs) # one dimensional representation
#     nA = config_.nA
#     [N,horizon,p] = X_.shape
#     K = config_.K
#     f = config_.arrival_type_dens
#     scalar_Ss = dict(zip(margSs, [ind for ind, s in enumerate(margSs)]))
#
#     # precompute the means
#     Xmean = np.mean(X,axis=0)
#
# """
# Generate  ALP constraints.
# """
#
#     # --------------------------------------------------------------------------
#     # Initialize to NULL the ALP constraint matrix, constraints right-hand-side,
#     # and objective coefficents.
#     constraintMatrix = None
#     RHSVector = None
#     objectiveVector = None
#
# # if loading from samples
#     # # --------------------------------------------------------------------------
#     # # Load sampled states, actions, and initial states.
#     # stateList = load(iPath + '/SAMPLED_STATE_ACTION_BASIS/stateList.npy')
#     # actionList = load(iPath + '/SAMPLED_STATE_ACTION_BASIS/actionList.npy')
#     # initList = load(iPath + '/SAMPLED_STATE_ACTION_BASIS/initialStateSamples.npy')
#     # ALPNumConstraint = len(stateList)
#
#
#     ALPNumConstraint = len(stateList)
# # --------------------------------------------------------------------------
#     # Auxilary redefinition of constants and functions.
#     discount = mdp.discount
#     evalBasisList = bf.evalBasisList
#     expectedBasisList = bf.expectedBasisList
#     getBatchOfNextStates = mdp.getBatchOfNextStates
#     getExpectedCost = mdp.getExpectedCost
#
# print(t)
# for a in range(nA):
#     print(a, 'a')
#     for s in margSs:
#         print(s)
#         # sum over sparse lookup
#         m.addConstrs((1.0 / N *
#                       gp.quicksum(
#                           np.abs(X_[i, t, p_])
#                           *  # average
#                           (
#                                   gp.quicksum(
#                                       f(t, k__) * X_[i, t, j] * theta[t, scalar_Ss[s], k__, j] for j in range(p) for k__
#                                       in range(K))
#                                   -
#                                   gp.quicksum(f(t, k) *
#                                               score(X_[i, t, :], A_[i, t], a, Y_[i, t], y, e_AX, Y_AX, K_[i, t], k) *
#                                               (R(s, a, sprime, k) +
#                                                gp.quicksum(f(t + 1, k__) *
#                                                            theta[t + 1, scalar_Ss[sprime], k__, j__] * np.mean(
#                                                    X_[t + 1, :, j__], axis=0)
#                                                            for j__ in range(p) for k__ in range(K))  # tilde V_t+1
#                                                )
#                                               for k in range(K) for (sprime, y) in G_s.map_[k][s]  # next states
#                                               )
#                           )
#                           for i in range(N)) >= 0 for p_ in range(p)), "alp")
#
#     # --------------------------------------------------------------------------
#     # Variable "X" has three components:
#     #   1) X[i][0]: vector of the value of basis functions list at the current
#     #               state.
#     #   2) X[i][1]: vector of the expected value of basis functions list at the
#     #               next state.
#     #   3) X[i][2]: ALP RHS values.
#     constraintMatrix = [X[i][0][:] - X[i][1][:] for i in range(ALPNumConstraint)]
#     RHSVector = [X[i][2] for i in range(ALPNumConstraint)]
#     expVFA = [X[i][1][:] for i in range(ALPNumConstraint)]
#
#     # --------------------------------------------------------------------------
#     # Save ALP constraints.
#     save(iPath + "/ALP_COMPONENTS/ALPConstMatrix", asarray(constraintMatrix), allow_pickle=True)
#     save(iPath + "/ALP_COMPONENTS/ALPRHSVector", asarray(RHSVector), allow_pickle=True)
#     save(iPath + "/ALP_COMPONENTS/ALPobjectiveVector", objectiveVector, allow_pickle=True)
#     save(iPath + "/ALP_COMPONENTS/exp_VFA", expVFA, allow_pickle=True)
#
#     return True
# def PLRALP_dynamic_Gs(config_,margSs, # horizon, p, K, J, M, nA, dS, margSs, Yy, # problem parameters
#             S_, X_, A_, Y_, e_AX, # data
#             G_s, R,
#            action_price_map, Y_AX,
#             score ):
#
#     # In this function, margSs refers to _marginal state margSs_
#
#     # dS is tuple of state dimension
#     # margSs is iterable of state variables, Yy is iterable of Y values
#     # dynamic_Gs assumes G_s is accessible in constant lookup
#     # X_ is [ N,T,p ] array
#     # assume score is a function of (X,A,Y,y,e_AX,Y_AX)
#     # load from config
#     dS = len(margSs) # one dimensional representation
#     nA = config_.nA
#     m = gp.Model()
#
#     # Create variables
#     # dimension of theta
#     [N,horizon,p] = X_.shape
#     theta_dim = tuple(horizon) + dS
#     theta = m.addMVar(theta_dim, vtype=GRB.CONTINUOUS, lb=0, name="theta")
#
#
#     # Set objective
#     X0mean = np.mean(X_[:,0,:],axis=0)
#     m.setObjective( gp.quicksum( X0mean @ theta[0,s] for s in range(dS)), gp.GRB.MINIMIZE)
#
#     for t in range(horizon):
#         for a in range(nA):
#             for s in margSs:
#                 # implement via sum over states multiplied by zero mapper
#                 # m.addConstr( gp.quicksum( np.abs(X_[i,t,0])*
#                 #             ((X_[i,t,:]@theta[t,s]) -
#                 #             gp.quicksum( score(X_[i,t,:],A_[i,t], Y_[i,t]) * G_s(s,sprime,y)*(R(s,a,sprime)+theta[t+1,sprime]@np.mean(X[t+1,:,:],axis=0) ) for sprime in margSs for y in Yy ) )
#                 #                                              for i in range(N) ), "alp")
#                 # sum over sparse lookup
#                 next_states = G_s.map_[s]
#                 # for state, y in next state
#                 for p_ in range(p):
#                     # different for multiple-product case
#                     m.addConstr( gp.quicksum( np.abs(X_[i,t,p_])*
#                             ((X_[i,t,:]@theta[t,s]) -
#                             gp.quicksum(  score(X_[i,t,:],A_[i,t], Y_[i,t],y,e_AX,Y_AX) *(R(s,a,sprime)+theta[t+1,sprime]@np.mean(X[t+1,:,:],axis=0) ) for (sprime,y) in next_states ) )
#                                                              for i in range(N) ), "alp")
#     # Optimize model
#     m.optimize()
#     return [m, theta]
